"""Server for multithreaded (asynchronous) chat application."""
import traceback
import asyncio
import websockets
import sys
from .player import Player


class Dungeon:

    name = 'Default Dungeon'
    description = 'Default description for a default dungeon... did you ' \
                  'forget to add a description?'

    server = None

    name2player = {}

    # the room where players start
    start_room = None

    host = None
    port = None

    def __init__(self, host='0.0.0.0', port=8080):
        self.host = host or '0.0.0.0'
        self.port = port or 8080
        self.init_dungeon()

    def init_dungeon(self):
        pass

    async def get_player_name(self, websocket):
        askname = 'Please enter a name less than 8 chars long'
        await websocket.send(askname)
        while True:
            name = await websocket.recv()
            name = name.strip()
            if name == '':
                # If an empty string is entered, don't say anything
                # also has the added benefit of a keepalive ping working
                # without generating duplicate prompts
                continue
            await websocket.send(askname)
            if name in self.name2player:
                await websocket.send(
                    'Name is already taken, please choose a name that is not '
                    'already chosen. Chosen names are: {}'.format(
                        '\n'.join(self.name2player.keys())
                    )
                )
            elif len(name) == 0:
                await websocket.send('Please do be so kind as to enter a name')
            elif len(name) >= 8:
                await websocket.send('Name is too long')
            else:
                return name

    async def accept_connection(self, websocket, *_):
        print('new connection, waiting for the name')
        name = await self.get_player_name(websocket)
        print('{} has arrived!'.format(name))
        player = Player(name)
        player.client = websocket
        player.tell('Welcome to {}, {}'.format(self.name, player.name))
        player.tell(self.description)
        player.tell('Hope your are ready to start your adventure! '
                    'Hit the enter key to enter the first room and then and '
                    'then call "describe_room" and "show_commands" to get '
                    'started!')
        (await websocket.recv()).strip()
        self.add_player(player)
        while True:
            try:
                cmd = (await websocket.recv()).strip()
                self.process_command(player, cmd)
            except websockets.exceptions.ConnectionClosed:
                self.remove_player(player)
                # breaking here will cause the connection to be closed
                break
            except KeyboardInterrupt as e:
                print(traceback.format_exc())
                break
            except Exception:
                print(traceback.format_exc())
                continue

    def start_dungeon(self):
        if not self.start_room:
            raise RuntimeError('No start room defined')
        print('Server is up and running on {}:{}'.format(self.host, self.port))
        start_server = websockets.serve(
            self.accept_connection, self.host, self.port, timeout=60 * 60
        )

        asyncio.get_event_loop().run_until_complete(start_server)
        asyncio.get_event_loop().run_forever()

    def add_player(self, player):
        self.name2player[player.name] = player
        self.start_room.add_player(player)

    def remove_player(self, player):
        print('removing', player.name)
        player.current_room.remove_player(player)
        del self.name2player[player.name]

    def broadcast(self, msg, name=None):
        if name is not None:
            msg = '{}: {}'.format(name, msg)
        for player in self.name2player.values():
            player.tell(msg)

    def process_command(self, player, raw_cmd):
        if raw_cmd.strip() == '':
            return
        cmd, args = raw_cmd.split()[0], raw_cmd.split()[1:]
        if cmd == 'say':
            self.tell_room(player, raw_cmd[4:])
        elif cmd == 'say_dungeon':
            self.broadcast(raw_cmd[12:], name=player.name)
        else:
            self.forward_cmd_to_room(player, cmd, args)

    def tell_room(self, player, msg):
        player.current_room.broadcast(msg, name=player.name)

    def forward_cmd_to_room(self, player, cmd, args):
        if cmd in player.current_room.commands:
            getattr(player.current_room, cmd)(player, *args)
        else:
            player.tell('{} not available in room {}'.format(
                cmd, player.current_room.name)
            )

    def kill_player(self, player):
        player.tell('You have died and are being returned to the beginning.')
        player.current_room.remove_player(player)
        self.start_room.add_player(player)




class Room:

    description = "This is a default room description. Be sure to come up " \
                  "with something better for this or your players will be " \
                  "bored AF."

    # name of the room
    name = 'Default Room'

    # The dungeon that the room belongs to
    dungeon = None

    # a mapping from player names to player objects
    players = None

    # a mapping of thing ids to things
    things = None

    # map from room id to the room object
    connected_rooms = None

    # an array of the commands that you want to open up to the players
    commands = None

    def __init__(self, dungeon):
        self.dungeon = dungeon
        self.players = {}
        self.things = {}
        self.connected_rooms = {}
        self.commands = [
            'say',
            'say_dungeon',
            'pick_up',
            'drop',
            'describe_room',
            'list_things',
            'describe_thing',
            'go_to',
            'show_commands',
        ]
        self.commands += self.register_commands()
        self.init_room()

    def init_room(self):
        pass

    def register_commands(self):
        return []

    def connect_room(self, other_room, two_way=True):
        self.connected_rooms[other_room.name] = other_room
        if two_way:
            other_room.connected_rooms[self.name] = self

    def disconnect_room(self, other_room, two_way=True):
        self.connected_rooms.pop(other_room.name, None)
        if two_way:
            other_room.connected_rooms.pop(self.name, None)

    def broadcast(self, message, name=None):
        if name is not None:
            message = '{}: {}'.format(name, message)
        for player in self.players.values():
            player.tell(message)

    def remove_player(self, player):
        del self.players[player.name]

    def add_player(self, player):
        self.players[player.name] = player
        player.tell('Welcome to the {} room'.format(self.name))
        # self.describe_room(player)
        # self.describe_room(player)
        player.current_room = self

    def pick_up(self, player, thing_id):
        if player.carrying:
            self.drop(player)

        # Returns True if successfully picked up, False otherwise
        if thing_id not in self.things:
            player.tell('No such thing with id {}'.format(thing_id))
            return False
        else:
            thing = self.things[thing_id]
            player.carrying = thing
            thing.carried_by = player
            del self.things[thing_id]
            player.tell('You have now picked up {}'.format(thing_id))
            self.after_pick_up(player)
            return True

    def pick_up_describe(self):
        return """Pick up a thing by its ID"""

    def drop(self, player):
        if player.carrying is not None:
            thing = player.carrying
            player.carrying = None
            thing.carried_by = None
            self.things[thing.id] = thing
            self.broadcast('{} dropped {}'.format(player.name, thing.id))
            self.after_drop(player, thing)
        else:
            player.tell('You aint carrying nothin')

    def drop_describe(self):
        return "Drop the item you are currently carrying"

    def describe_room(self, player):
        desc_str = self.description
        desc_str += '\nContains {} players and {} items\n'.format(
            len(self.players), len(self.things)
        )
        # desc_str += '\nThings in room:\n'
        # for thing in self.things.values():
        #     desc_str += thing.id + ' ' + thing.description + '\n'
        if self.connected_rooms:
            desc_str += 'This room has doors to the following rooms:\n'
            for room_id in self.connected_rooms:
                desc_str += '{}\n'.format(room_id)
        else:
            desc_str += 'This room has no active doors to other rooms\n'
        player.tell(desc_str)

    def describe_room_describe(self):
        return """Describe all objects, things"""

    def list_things(self, player):
        msg = 'Things:\n'
        for thing in self.things:
            msg += '{}\n'.format(thing)
        player.tell(msg)

    def list_things_describe(self):
        return 'List all things that may be picked up in the room'

    def describe_thing(self, player, thing_id):
        player.tell(self.things[thing_id].long_description())

    def describe_thing_describe(self):
        return "List all things that are in the room"

    def go_to(self, player, other_room_name):
        if other_room_name not in self.connected_rooms:
            player.tell(
                "There is no connected room named {}".format(other_room_name))

        else:
            self.remove_player(player)
            self.connected_rooms[other_room_name].add_player(player)

    def go_to_describe(self):
        return "Select a room to leave to"

    def show_commands(self, player):
        desc_str = 'Room Commands:\n'
        for command in self.commands:
            desc_str += '{} - {}\n'.format(
                command,
                getattr(self, command + '_describe')(),
            )
        player.tell(desc_str)

    def show_commands_describe(self):
        return "Show all commands that are available to the room"

    def say_describe(self):
        return 'Say something to everyone in the room you are currently in'

    def say_dungeon_describe(self):
        return 'Say something to the entire Dungeon (all people in all rooms)'

    def after_pick_up(self, player):
        # called after a user pick something up
        pass

    def after_drop(self, player, thing):
        # called after a user drops something
        pass

    def add_thing(self, thing):
        self.things[thing.id] = thing




# global counter for things
count = 0


class Thing:

    id = None
    name = 'Thing'
    description = 'you need to describe the thing'
    # the player currently holding the thing
    held_by = None

    def __init__(self):
        global count
        count += 1
        self.id = '{}_{}'.format(self.name, count)

    def picked_up_by(self, player):
        pass

    def dropped_by(self, player):
        pass



import asyncio


class Player:

    # string name, guaranteed to be unique
    name = None

    # the room that the player is currently in
    current_room = None

    # the python websocket through which you can communicate
    client = None

    # the clothing that the player is currently wearing
    clothing = None

    # what the user is currently carrying
    carrying = None

    def __init__(self, name):
        self.name = name

    async def _tell(self, msg):
        await self.client.send(msg + '\n')

    def tell(self, msg):
        asyncio.ensure_future(self._tell(msg))



from setuptools import setup

setup(
    name='novamud',
    version='0.0.8',
    description='Simple and flexible mud framework for people with only basic '
                'knowledge of python',
    author='Sam Hopkins',
    author_email='sam@daredata.engineering',
    url='https://gitlab.com/hershaw/novamud',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    packages=['novamud'],
    install_requires=[
        'websockets==6.0'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
    ]
)
